from torcherist.nlp import *
from torcherist.dataset.text import read_ptb
import pytest
from itertools import tee

doc = read_ptb()
vocab = Vocab(doc)


def test_count_text():
    counter = count_text(doc)
    assert counter["the"] == 50770


def test_Vocab():
    assert len(vocab) == 6719
    with pytest.raises(KeyError):
        vocab.token2idx["aer"]

    # aer rare word, map to unk
    assert vocab[["<unk>", "the", "aer"]] == [0, 1, 0]
    assert vocab.to_token([[0, 1], [0]]) == [["<unk>", "the"], ["<unk>"]]
    assert vocab.get_counts("the") == 50770
    assert vocab.get_counts("aer") == 1
    for sent in vocab.lookup(doc):
        assert "aer" not in sent
    corpus_gen = (token for sent in vocab.doc2corpus(doc) for token in sent)
    g1, g2 = tee(corpus_gen, 2)
    assert min(g1) == 0
    assert max(g2) == len(vocab) - 1



import torch
import torch.nn as nn
from torcherist.tensor import *

def test_sequence_mask():
    # X = torch.tensor([[1, 2, 3], [4, 5, 6]])
    valid_len = torch.tensor([1,3,2])
    assert torch.all(torch.eq(
        sequence_mask(torch.tensor([1,3,2]), 5),
        torch.tensor(
            [[True, False, False, False, False],
             [True, True, True, False, False],
             [True, True, False, False, False]], dtype=torch.bool
        ))
    )

    assert torch.all(torch.eq(
        sequence_mask(torch.tensor([[1,3],[2,0]])),
        torch.tensor(
            [[[True, False, False],
              [True, True, True]],
             [[True, True, False],
              [False, False, False]]], dtype=torch.bool
        )
    ))

    # masked softmax cross entropy
    pred = torch.ones(3, 4, 10)
    label = torch.ones((3, 4), dtype=torch.long)
    valid_len = torch.tensor([4, 2, 0])
    weights = torch.ones_like(label)
    # fill irrelevant items in weights to 0
    fill_sequence_mask(weights, valid_len)

    unweighted_loss = nn.CrossEntropyLoss(reduction="none")(
                    pred.permute(0, 2, 1), label)
    weighted_loss = (unweighted_loss * weights).mean(dim=1)
    assert torch.allclose(weighted_loss, torch.tensor([2.3026, 1.1513, 0.000]))



import random

from torcherist.nlp import Corpus, flatten_text, Vocab
from typing import List
import random
from torcherist.dataset.text import read_time_machine
import torch

def sequential_text_sampler(corpus: List[int],
                            by="random",
                            batch_size: int = 32,
                            m: int = 5):
    """
    sample sequential text by two strategy

    Args:
        corpus: list of index (should not be str), and should be flatten from Corpus
        by: strategy, either "random" or "sequential"
        m: integer, max len of each sample in one batch

    Returns:
        X: (batch_size, m), in total `num_subseqs` // `batch_size` number of batches
        y: (batch_size, m) y_ij is X_ij's next token

    """
    assert by in ["random", "sequential"], 'by must be either "random" or "sequential"'

    if by == "random":
        # Start with a random offset to partition a sequence
        # ignore first few tokens
        corpus = corpus[random.randint(0, m):]
        # Subtract 1 since we need to account for labels
        num_subseqs = (len(corpus) - 1) // m
        # The starting indices for subsequences of length `m`
        initial_indices = list(range(0, num_subseqs * m, m))
        # In random sampling, the subsequences from two adjacent random
        # minibatches during iteration are not necessarily adjacent on the
        # original sequence
        random.shuffle(initial_indices)

        def data(pos):
            # Return a sequence of length `m` starting from `pos`
            return corpus[pos: pos + m]

        num_batches = num_subseqs // batch_size
        for i in range(0, batch_size * num_batches, batch_size):
            # Here, `initial_indices` contains randomized starting indices for
            # subsequences
            initial_indices_per_batch = initial_indices[i: i + batch_size]
            X = [data(j) for j in initial_indices_per_batch]
            y = [data(j + 1) for j in initial_indices_per_batch]
            yield torch.tensor(X), torch.tensor(y)

    # sequential partition
    else:
        # start with a random offset (ignore first few tokens)
        offset = random.randint(0, m)
        # total number of tokens in all return X
        # Subtract 1 since we need to account for labels
        num_tokens = ((len(corpus) - offset - 1) // batch_size) * batch_size
        Xs = torch.tensor(corpus[offset: offset + num_tokens]).reshape(batch_size, -1)
        ys = torch.tensor(corpus[offset + 1: offset + 1 + num_tokens]).reshape(batch_size, -1)
        num_batches = Xs.shape[1] // m
        for i in range(0, num_batches * m, m):
            # ensure next batch is next to cur batch (sequential)
            yield Xs[:, i: i + m], ys[:, i: i + m]


def load_time_machine(batch_size: int, m: int,
                      by: str,
                      max_tokens = 10000,
                      *,
                      filepath: str = None):
    """
    load time machine text in character levels

    Args:
        batch_size: batch size
        m: generate batch of m-len characters
        by: sample strategy
        max_tokens: only load `max_tokens` chars from time machine
        filepath: path of time machine text

    Returns:
        data_iter: TimeMachineLoader, custom __iter__ so that can loop through data multiple time
        vocab
    """
    class TimeMachineLoader:
        def __init__(self, iter_fn: callable, *arg):
            self.iter_fn = iter_fn
            self. arg = arg
        def __iter__(self):
            return self.iter_fn(*self.arg)

    if filepath:
        doc = read_time_machine(filepath)
    else:
        doc = read_time_machine()

    # map to character level
    #  list of (list of char / word)
    doc = [list(token) for sent in doc for token in sent]
    vocab = Vocab(doc)

    corpus = list(flatten_text(vocab.doc2corpus(doc)))
    corpus = corpus[:max_tokens]

    return TimeMachineLoader(
        sequential_text_sampler,
        corpus,
        by,
        batch_size,
        m
    ), vocab



import time

import numpy as np
import torch


class Timer:
    """Record multiple running times."""

    def __init__(self):
        self.times = []
        self.start()

    def start(self):
        """Start the timer."""
        self.tik = time.time()

    def stop(self):
        """Stop the timer and record the time in a list."""
        self.times.append(time.time() - self.tik)
        return self.times[-1]

    def avg(self):
        """Return the average time."""
        return sum(self.times) / len(self.times)

    def sum(self):
        """Return the sum of time."""
        return sum(self.times)

    def cumsum(self):
        """Return the accumulated time."""
        return np.array(self.times).cumsum().tolist()


class Benchmark:
    """context manager that print time

    with Benchmark("desc"):
        X @ W + b
    """
    def __init__(self, description='Done'):
        self.description = description

    def __enter__(self):
        self.timer = Timer()
        return self

    def __exit__(self, *args):
        print(f'{self.description}: {self.timer.stop():.4f} sec')


def try_gpu(use_cpu = False):
    if use_cpu:
        return torch.device("cpu")
    return torch.device('cuda' if torch.cuda.is_available() else 'cpu')



import matplotlib.pyplot as plt
from IPython import display


class Accumulator:
    """For accumulating sums over `n` variables."""
    def __init__(self, n):
        self.data = [0.0] * n

    def add(self, *args):
        self.data = [a + float(b) for a, b in zip(self.data, args)]

    def reset(self):
        self.data = [0.0] * len(self.data)

    def __getitem__(self, idx):
        return self.data[idx]


class Animator:
    @staticmethod
    def use_svg_display():
        """Use the svg format to display a plot in Jupyter."""
        display.set_matplotlib_formats('svg')

    @staticmethod
    def set_axes(axes, xlabel, ylabel, xlim, ylim, xscale, yscale, legend):
        """Set the axes for matplotlib."""
        axes.set_xlabel(xlabel)
        axes.set_ylabel(ylabel)
        axes.set_xscale(xscale)
        axes.set_yscale(yscale)
        axes.set_xlim(xlim)
        axes.set_ylim(ylim)
        if legend:
            axes.legend(legend)
        axes.grid()

    """For plotting data in animation."""

    def __init__(self, xlabel=None, ylabel=None, legend=None, xlim=None,
                 ylim=None, xscale='linear', yscale='linear',
                 fmts=('-', 'm--', 'g-.', 'r:'), nrows=1, ncols=1,
                 figsize=(3.5, 2.5)):

        # Incrementally plot multiple lines
        if legend is None:
            legend = []
        Animator.use_svg_display()
        self.fig, self.axes = plt.subplots(nrows, ncols, figsize=figsize)
        if nrows * ncols == 1:
            self.axes = [self.axes, ]
        # Use a lambda function to capture arguments
        self.config_axes = lambda: Animator.set_axes(
            self.axes[0], xlabel, ylabel, xlim, ylim, xscale, yscale, legend)
        self.X, self.Y, self.fmts = None, None, fmts

    def add(self, x, y):
        # Add multiple data points into the figure
        if not hasattr(y, "__len__"):
            y = [y]
        n = len(y)
        if not hasattr(x, "__len__"):
            x = [x] * n
        if not self.X:
            self.X = [[] for _ in range(n)]
        if not self.Y:
            self.Y = [[] for _ in range(n)]
        for i, (a, b) in enumerate(zip(x, y)):
            if a is not None and b is not None:
                self.X[i].append(a)
                self.Y[i].append(b)
        self.axes[0].cla()
        for x, y, fmt in zip(self.X, self.Y, self.fmts):
            self.axes[0].plot(x, y, fmt)
        self.config_axes()
        display.display(self.fig)
        display.clear_output(wait=True)



"""Use exact matching with a source list to find skills"""

import logging

import nltk
try:
    nltk.sent_tokenize('test')
except LookupError:
    nltk.download('punkt')

from .base import (
    CandidateSkill,
    ListBasedSkillExtractor,
    CandidateSkillYielder,
    trie_regex_from_words
)

from typing import Dict


class ExactMatchSkillExtractor(ListBasedSkillExtractor):
    """Extract skills from unstructured text

    Builds a lookup based on the 'name' attribute of all competencies in the given framework

    Originally written by Kwame Porter Robinson
    """
    method_name = 'exact_match'
    method_description = 'Exact matching'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        competencies = set(
            competency
            for competency in self.competency_framework.values()
            if competency.name
        )

        self.id_lookup = dict((competency.name.lower(), [competency.identifier, competency.categories]) for competency in competencies)

        logging.info(
            'Found %s entries for lookup',
            len(competencies)
        )
        self.lookup_regex = trie_regex_from_words(set(self.id_lookup.keys()))

    def _skills_lookup(self) -> set:
        """Create skills lookup

        Read names from Ontology into a set

        Returns: (set) skill names
        """

    def candidate_skills(self, source_object: Dict) -> CandidateSkillYielder:
        """Yield objects which may represent skills from the given source object.

        Looks for exact matches between the reference skill lookup and the object's text.

        Args: source_object (dict) A structured document for searching, such as a job posting

        Yields: CandidateSkill objects
        """
        document = self.transform_func(source_object)
        sentences = self.nlp.sentence_tokenize(document, include_spans=True)
        for sent in sentences:
            matches = self.lookup_regex.finditer(sent.text)
            for match in matches: 
                logging.info('Yielding exact match %s in string %s', match, sent.text)
                yield CandidateSkill(
                    skill_name=match[0].lower(),
                    matched_skill_identifier=self.id_lookup[match[0].lower()][0],
                    skill_cat=self.id_lookup[match[0].lower()][1],
                    confidence=100,
                    context=sent.text,
                    start_index=sent.start_index + match.start(),
                    document_id=source_object['id'],
                    document_type=source_object['@type'],
                    source_object=source_object,
                    skill_extractor_name=self.name
                )



from .base import SkillExtractor, CandidateSkill, CandidateSkillYielder
from ..nlp import section_extract
import logging
from typing import Dict


class SectionExtractSkillExtractor(SkillExtractor):
    """Extract skills from text by extracting sentences from matching 'sections'.

    Heavily utilizes skills_ml.algorithms.nlp.section_extract.
    For more detail on how to define 'sections', refer to its docstring.
    """
    def __init__(self, section_regex=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.section_regex = section_regex or r'.*([Qq]ualifications|[Ss]kills|[Rr]equirements|[Ab]ilities|[Cc]ompetencies)'

    @property
    def name(self):
        return f'section_extract_{self.section_regex}'

    @property
    def description(self):
        return f'Sentences from section matching regular expression: {self.section_regex}'

    def candidate_skills(self, source_object: Dict) -> CandidateSkillYielder:
        """Generate candidate skills from the source object

        Yields each sentence from the configured section pattern
        """

        spans_in_section = section_extract(self.section_regex, source_object['description'])
        for span in spans_in_section:
            logging.info('Yielding candidate skill %s', span)
            yield CandidateSkill(
                skill_name=span.text,
                matched_skill_identifier=None,
                confidence=100,
                context=span.text,
                start_index=span.start_index,
                document_id=source_object['id'],
                document_type=source_object['@type'],
                source_object=source_object,
                skill_extractor_name=self.name
            )



#!/usr/bin/env python

from setuptools import find_packages, setup

setup(
    name='Torcherist',
    version='0.1.6',
    description='utils codes for pyTorch',
    author="Jiashu Xu",
    author_email='1999J0615une@gmail.com',
    url='https://github.com/cnut1648/torcher',
    download_url='https://github.com/cnut1648/torcherist/archive/v0.1.2.tar.gz',
    packages=find_packages(),
    license="MIT license",
    keywords=['nlp', 'skills', 'onet', 'pytorch'],
    install_requires=[
        'beautifulsoup4',
        'numpy',
        'graphviz',
        'matplotlib',
        'mosestokenizer',
        'requests',
        'nltk',
        'torch',
        'cached_property'
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
    ],
)



"""Base classes for skill extraction"""
import logging
from abc import ABCMeta, abstractmethod

from collections import Counter, namedtuple


from torcherist.algorithms.skill_extractors.corpus import SimpleCorpusCreator

from .. import nlp
from torcherist.dataset.ontology import CompetencyFramework
# from skills_ml.ontologies.base import CompetencyFramework

from typing import Dict, Callable, Text, Generator
import re


class Trie():
    """Regex::Trie in Python. Creates a Trie out of a list of words. The trie can be exported to a Regex pattern.
    The corresponding Regex should match much faster than a simple Regex union."""

    def __init__(self):
        self.data = {}

    def add(self, word):
        ref = self.data
        for char in word:
            ref[char] = char in ref and ref[char] or {}
            ref = ref[char]
        ref[''] = 1

    def dump(self):
        return self.data

    def quote(self, char):
        return re.escape(char)

    def _pattern(self, pData):
        data = pData
        if "" in data and len(data.keys()) == 1:
            return None

        alt = []
        cc = []
        q = 0
        for char in sorted(data.keys()):
            if isinstance(data[char], dict):
                try:
                    recurse = self._pattern(data[char])
                    alt.append(self.quote(char) + recurse)
                except:
                    cc.append(self.quote(char))
            else:
                q = 1
        cconly = not len(alt) > 0

        if len(cc) > 0:
            if len(cc) == 1:
                alt.append(cc[0])
            else:
                alt.append('[' + ''.join(cc) + ']')

        if len(alt) == 1:
            result = alt[0]
        else:
            result = "(?:" + "|".join(alt) + ")"

        if q:
            if cconly:
                result += "?"
            else:
                result = "(?:%s)?" % result
        return result

    def pattern(self):
        return self._pattern(self.dump())


def trie_regex_from_words(words):
    trie = Trie()
    for word in words:
        trie.add(word)
    return re.compile(r"\b" + trie.pattern() + r"\b", re.IGNORECASE)


CandidateSkill = namedtuple('CandidateSkill', [
    'skill_name',
    'matched_skill_identifier',
    'skill_cat',
    'context',
    'start_index',
    'confidence',
    'document_id',
    'document_type',
    'source_object',
    'skill_extractor_name'
])


CandidateSkillYielder = Generator[CandidateSkill, None, None]


class SkillExtractor(object, metaclass=ABCMeta):
    """Abstract class for all skill extractors.

    All subclasses must implement candidate_skills.

    All subclasses must define properties
    'method' (a short machine readable property)
    'description' (a text description of how the extractor does its work)

    Args:
        transform_func (callable, optional) Function that transforms a structured object into text
            Defaults to SimpleCorpusCreator's _join, which takes common text fields
            in common schema job postings and concatenates them together.
            For non-job postings another transform function may be needed.
    """
    def __init__(self, transform_func: Callable=None):
        self.transform_func = transform_func
        if not self.transform_func:
            self.transform_func = SimpleCorpusCreator()._join
        self.nlp = nlp

    @property
    @abstractmethod
    def name(self):
        """A short, machine-friendly (ideally snake_case) name for the skill extractor"""
        pass

    @property
    @abstractmethod
    def description(self):
        """A human-readable description for the skill extractor"""
        pass

    @abstractmethod
    def candidate_skills(self, source_object: Dict) -> CandidateSkillYielder:
        """Yield objects which may represent skills/competencies from the given source object

        Args: source_object (dict) A structured document for searching, such as a job posting

        Yields: CandidateSkill objects
        """
        pass

    def document_skill_counts(self, source_object: Dict):
        """Count skills in the document

        Args:
            source_object (dict) A structured document for searching, such as a job posting

        Returns: (collections.Counter) skills found in the document, all
            values set to 1 (multiple occurrences of a skill do not count)
        """
        skill_counts = Counter()
        for candidate_skill in self.candidate_skills(source_object):
            skill_counts[self.nlp.lowercase_strip_punc(candidate_skill.skill_name).lstrip().rstrip()] += 1
        return skill_counts


class ListBasedSkillExtractor(SkillExtractor):
    """Extract skills by comparing with a known lookup/list.

    Subclasses must implement _skills_lookup and _document_skills_in_lookup

    Args:
        skill_lookup_name (string, optional) An identifier for the skill lookup type.
            Defaults to onet_ksat
        skill_lookup_description (string, optional) A human-readable description of the skill lookup.
    """
    def __init__(self, competency_framework, *args, **kwargs):
        super(ListBasedSkillExtractor, self).__init__(*args, **kwargs)
        if not isinstance(competency_framework, CompetencyFramework):
            raise ValueError('Must pass in a CompetencyFramework object')
        if not competency_framework.name or not competency_framework.description:
            raise ValueError('CompetencyFramework object must be documented with a name and description')
        self.competency_framework = competency_framework

    @property
    @abstractmethod
    def method_name(self):
        """A short, machine-friendly name for the method of skill extraction, independent of the skill lookup"""
        pass

    @property
    @abstractmethod
    def method_description(self):
        """A human readable description for the method of skill extraction, independent of the skill lookup"""
        pass

    @property
    def name(self):
        return f'{self.competency_framework.name}_{self.method_name}'

    @property
    def description(self):
        return f'{self.competency_framework.description} found by {self.method_description}'



"""String transformations for cleaning
for unicodedata, see:
http://www.unicode.org/reports/tr44/tr44-4.html#General_Category_Values
"""
import unicodedata
import re
from bs4 import BeautifulSoup
import nltk
from nltk.tokenize import PunktSentenceTokenizer
from functools import reduce, wraps
from typing import List, Set, Generator, Dict, Pattern, Tuple
from collections import namedtuple


transforms = ['nlp_a']
BULLET_CHARACTERS = ['+', '*', '-']


def deep(func):
    """A decorator that will apply a function to a nested list recursively

    Args:
        func (function): a function to be applied to a nested list
    Returns:
        function: The wrapped function
    """
    @wraps(func)
    def wrapper(*args, **kwargs):
        if isinstance(args[0], list):
            if len(args[0]) == 1:
                return wrapper(args[0][0])
            return [wrapper(i) for i in args[0]]
        return func(args[0])
    return wrapper


def normalize(text: str) -> str:
    """
    Args:
        text (str): A unicode string
    Returns:
        str: The text, lowercased and in NFKD normal form
    """
    return unicodedata.normalize('NFKD', text.lower())

@deep
def clean_html(text: str) -> str:
    markup = BeautifulSoup(text, "lxml")
    return unicodedata.normalize('NFKD', markup.get_text())


def lowercase_strip_punc(text: str, punct: Set[str]=None) -> str:
    """
    Args:
        text (str): A unicode string
        punct (:obj: `set`, optional)
    Returns:
        str: The text, lowercased, sans  punctuation and in NFKD normal form
    """
    if not punct:
        punct = set(['P', 'S'])

    return ''.join(
        char for char in normalize(text)
        if not unicodedata.category(char)[0] in punct
    )


def title_phase_one(text: str, punct: Set[str]=None) -> str:
    """
    Args:
        text (str): A unicode string
        punct (:obj: `set`, optional)
    Returns:
        str: The text, lowercased, sans punctuation, whitespace normalized
    """
    if not punct:
        punct = set(['P', 'S'])
    no_apos = re.sub(r'\'', '', normalize(text))
    strip_punc = ''.join(
        char if not unicodedata.category(char)[0] in punct else ' '
        for char in no_apos
    )
    return re.sub(r'\s+', ' ', strip_punc.strip())

@deep
def clean_str(text: str) -> str:
    """
    Args:
        text: A unicode string
    Returns:
        str: lowercased, sans punctuation, non-English letters
    """
    RE_PREPROCESS = r'\W+|\d+'
    text = re.sub(
        RE_PREPROCESS,
        ' ',
        text.lower()
    )
    text = re.sub(r"[^A-Za-z0-9(),!?\'\`]", " ", text)
    text = re.sub(r"\'s", " \'s", text)
    text = re.sub(r"\'ve", " \'ve", text)
    text = re.sub(r"n\'t", " n\'t", text)
    text = re.sub(r"\'re", " \'re", text)
    text = re.sub(r"\'d", " \'d", text)
    text = re.sub(r"\'ll", " \'ll", text)
    text = re.sub(r"\s{2,}", " ", text)
    return text


Span = namedtuple("Span", ["text", "start_index"])

def sentence_tokenize(text: str, include_spans: bool=False) -> List:
    """
    Args:
        text (str): a unicode string
    Returns:
        list: tokenized sentence. If include_spans is True, then each item is a Span object  with both text and a start_index. Otherwise, only text is returned.
    """
    lines = re.split('\n', text)
    #lines = list(filter(None, sentences))
    sentences = []
    
    tokenizer = PunktSentenceTokenizer()
    total_offset = 0
    for line in lines:
        line_start = total_offset
        for sentence_start, sentence_end in tokenizer.span_tokenize(line):
            sentence = line[sentence_start:sentence_end]
            if include_spans:
                sentences.append(Span(text=sentence, start_index=line_start + sentence_start))
            else:
                sentences.append(sentence)
            total_offset = line_start + sentence_end
        total_offset += 1
    return sentences

@deep
def word_tokenize(text: str, punctuation=True) -> List[str]:
    """
    Args:
        text (str): a unicode string
    Returns:
        list: tokenized words
    """
    if punctuation:
        return nltk.word_tokenize(text)
    else:
        return nltk.wordpunct_tokenize(text)


def fields_join(
        document: Dict,
        document_schema_fields: List[str]=None) -> str:
    """
    Args:
        document (dict): a document dictionary
        document_schema_fields (:obj: `list`, optional): a list of keys
    Returns:
        str: a text joined with selected fields.
    """
    if not document_schema_fields:
        document_schema_fields = [
                'description',
                'experienceRequirements',
                'qualifications',
                'skills']
    return ' '.join([document.get(field, '') for field in document_schema_fields])


def vectorize(
        tokenized_text: List[str],
        embedding_model):
    """
    Args:
        tokenized_text: a tokenized list of word tokens
        embedding_model: the embedding model implements `.infer_vector()` method
    Returns:
        np.ndarray: a word embedding vector
    """

    return embedding_model.infer_vector(tokenized_text)


def section_extract(section_regex: Pattern, document: str) -> List[Span]:
    """Only return the contents of the configured section heading

    Defines a 'heading' as the text of a sentence that:
        - does not itself start with a bullet character
        - either has between 1 and 3 words or ends in a colon

    For a heading that matches the given pattern, returns each sentence between it and the next heading.

    Heavily relies on the fact that sentence_tokenize does line splitting
    as well as standard sentence tokenization. In this way, it should work both
    for text strings that have newlines and for text strings that don't.

    In addition, this function splits each sentence by bullet characters as often bullets denote
    what we want to call 'sentences', but authors often take advantage of the bullet characters
    to make the contents of each 'sentence' into small sentence fragments, which makes standard
    sentence tokenization insufficient if the newlines have been taken out.

    Args:
        section_regex (Pattern), A regular expression defining the heading/s you want to include
        document (str) The text to search in
    Returns: List of Span objects with both the text and a start_index
    """
    units_in_section = []
    if not document:
        return units_in_section
    sentences = sentence_tokenize(document, include_spans=True)
    units = [
        Span(text=unit.text, start_index=sentence.start_index + unit.start_index)
        for sentence in sentences
        for unit in split_by_bullets(sentence.text)
    ]

    heading = ''
    for unit in units:
        words_in_unit = len(unit.text.lstrip().rstrip().split(' '))
        if unit.text.strip() and unit.text[0] not in BULLET_CHARACTERS and ((words_in_unit > 0 and words_in_unit < 4) or unit.text.endswith(':')):
            heading = unit.text
        if re.match(section_regex, heading) and unit.text != heading and len(unit.text.strip()) > 0:
            stripped = strip_bullets_from_line(unit.text).lstrip().rstrip()

            units_in_section.append(Span(
                text=stripped,
                start_index=unit.start_index + unit.text.index(stripped)
            ))
    return units_in_section


def split_by_bullets(sentence: str) -> List[Span]:
    """Split sentence by bullet characters

    Args:
        sentence (str)

    Returns: List of Span objects representing the text inbetween bullets, with both text and start indices
    """
    units = []
    for bullet_char in BULLET_CHARACTERS:
        index = 0
        padded_bullet = bullet_char + ' '
        if sentence.count(padded_bullet) > 1:
            for i, fragment in enumerate(sentence.split(padded_bullet)):
                if i > 0:
                    units.append(Span(text=padded_bullet + fragment, start_index=index))
                    index += len(padded_bullet + fragment)
                else:
                    units.append(Span(text=fragment, start_index=index))
                    index += len(fragment)
            return units
    units.append(Span(text=sentence, start_index=0))
    return units


def strip_bullets_from_line(line: str) -> str:
    """Remove bullets from beginning of line"""
    for bullet_char in BULLET_CHARACTERS:
        if line.startswith(bullet_char):
            line = line.replace(bullet_char, '')
    return line
